#include <avr/io.h>
#include <avr/twi.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <stdio.h>

#include "uart.h"

#define STA015_TWI_ADDR		0x43

uint8_t twiReceive(uint8_t ack) {
	uint8_t twcr = (1<<TWINT) | (1<<TWEN);
	if(ack) twcr |= (1<<TWEA);
	// clear interrupt to start transmission
	TWCR = twcr;
	// wait for transmission
	while ((TWCR & (1<<TWINT)) == 0);
	// return data
	return TWDR;
}

uint8_t twiTransmit(uint8_t data) {
	// transmit data
	TWDR = data;
	// clear interrupt to start transmission
	TWCR = (1<<TWINT) | (1<<TWEN);
	// wait for transmission
	while ((TWCR & (1<<TWINT)) == 0);
	// return status code
	return TW_STATUS;
}

uint8_t twiStart(void) {
	// send start condition
	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
	// wait for transmission
	while ((TWCR & (1<<TWINT)) == 0);
	// return status code
	return TW_STATUS;
}

void twiStop(void) {
	// send stop condition
	TWCR = (1<<TWINT) | (1<<TWSTO) | (1<<TWEN);
}

uint8_t sta015WriteRegister(uint8_t addr, uint8_t value) {
	uint8_t status, error=0;
	
	//delay
	for(volatile uint16_t i=0; i<200; i++);
	
	// start condition
	status = twiStart();
	if(status != TW_REP_START && status != TW_START) {
		// Error, keine Stop Condition senden
		printf_P(PSTR("SC Error: 0x%.2X\r"), status);
		return -1;
	}
	
	// send SLA+W
	status = twiTransmit((STA015_TWI_ADDR<<1) | TW_WRITE);
	if(status != TW_MT_SLA_ACK) {
		// Error (wahrscheinlich ein NACK...)
		printf_P(PSTR("SLA+W Error: 0x%.2X\r"), status);
		error = status;
		goto quit;
	}
	
	// send register address
	status = twiTransmit(addr);
	if(status != TW_MT_DATA_ACK) {
		// Error
		printf_P(PSTR("Reg Error: 0x%.2X\r"), status);
		error = status;
		goto quit;
	}
	
	// write value
	status = twiTransmit(value);
	if(status != TW_MT_DATA_ACK) {
		// Error
		printf_P(PSTR("Write Error: 0x%.2X\r"), status);
		error = status;
	}

quit:
	// Stop condition
	twiStop();
	
	return error;
}

int16_t sta015ReadRegister(uint8_t addr) {
	uint8_t status;
	int16_t data=0;
	
	// start condition
	status = twiStart();
	if(status != TW_REP_START && status != TW_START) {
		// Error, keine Stop Condition senden
		printf_P(PSTR("SC Error: 0x%.2X\r"), status);
		return -1;
	}
	
	// send SLA+W
	status = twiTransmit((STA015_TWI_ADDR<<1) | TW_WRITE);
	if(status != TW_MT_SLA_ACK) {
		// Error (wahrscheinlich ein NACK...)
		printf_P(PSTR("SLA+W Error: 0x%.2X\r"), status);
		data = -1;
		goto quit;
	}
	
	// send register address
	status = twiTransmit(addr);
	if(status != TW_MT_DATA_ACK) {
		// Error
		printf_P(PSTR("Reg Error: 0x%.2X\r"), status);
		data = -1;
		goto quit;
	}
	
	// repeated start condition
	status = twiStart();
	if(status != TW_REP_START && status != TW_START) {
		// Error, keine Stop Condition senden
		printf_P(PSTR("RSC Error: 0x%.2X\r"), status);
		return -1;
	}
	
	// send SLA+R
	status = twiTransmit((STA015_TWI_ADDR<<1) | TW_READ);
	if(status != TW_MR_SLA_ACK) {
		// Error (wahrscheinlich ein NACK...)
		printf_P(PSTR("SLA+R Error: 0x%.2X\r"), status);
		data = -1;
		goto quit;
	}
	
	// register lesen
	data = twiReceive(0);
	// hier gibs kein Error-Handling, aber wenn bis hierher alles
	// funktioniert hat, wirds schon richtig sein :P

quit:
	// Stop condition
	twiStop();
	
	return data;
}

int main(void) {
	uint8_t buffer[512];
	
	// SPI initialisieren
	DDRB |= (1<<PB0) | (1<<PB1) | (1<<PB2) | (1<<PB7);
	SPCR = (1<<SPE) | (1<<MSTR) | (1<<CPHA);
	
	// UART initialisieren
	uartInit();
	puts_P(PSTR("STA015 Eval Board\r\r"));
	
	// I2C initialisieren
	TWBR = 20;
	
	// STA015 initialisieren...
	puts_P(PSTR("STA015 initialisieren...\r"));
	
	sta015WriteRegister(0x10, 0x01);	// SOFT_RESET
	// ein bisschen warten
	for(volatile uint16_t w=0; w<1000; w++);
	
	// IDENT lesen (=0xAC)
	int16_t data;
	data = sta015ReadRegister(0x01);
	printf_P(PSTR("IDENT: 0x%.2X\r"), data);
	// VERSION lesen
	data = sta015ReadRegister(0x00);
	printf_P(PSTR("VERSION: 0x%.2X\r"), data);
	// SOFTVERSION lesen
	data = sta015ReadRegister(0x71);
	printf_P(PSTR("SOFTVERSION: 0x%.2X\r"), data);
	// HEAD lesen
	
	puts_P(PSTR("konfigurieren... "));
	sta015WriteRegister(0x54, 0x03);	// PCM_DIVIDER
	sta015WriteRegister(0x55, 0x20);	// PCMCONF
	sta015WriteRegister(0x07, 0x00);	// PLLCTL [15:12]
	sta015WriteRegister(0x06, 0x16);	// PLLCTL [20:16]
	sta015WriteRegister(0x0B, 0x03);	// ???
	sta015WriteRegister(0x52, 0xFD);	// PLLFRAC_441_H
	sta015WriteRegister(0x51, 0x8B);	// PLLFRAC_441_L
	sta015WriteRegister(0x65, 0x0A);	// PLLFRAC_H
	sta015WriteRegister(0x64, 0x3D);	// PLLFRAC_L
	sta015WriteRegister(0x50, 0x10);	// MFSDF_441
	sta015WriteRegister(0x61, 0x0E);	// MSDF(X)
	sta015WriteRegister(0x05, 0xA1);	// PLLCTL [7:0]
	sta015WriteRegister(0x0D, 0x04);	// SCLK_POL
	sta015WriteRegister(0x4D, 0x00);	// CHIP_MODE
	sta015WriteRegister(0x18, 0x04);	// DATA_REQ_ENABLE
	sta015WriteRegister(0x0C, 0x01);	// REQ_POL
	sta015WriteRegister(0x14, 0x01);	// MUTE
	sta015WriteRegister(0x78, 0x0B);	// TREBLE_FREQUENCY_HIGH
	sta015WriteRegister(0x77, 0xB8);	// TREBLE_FREQUENCY_LOW
	sta015WriteRegister(0x7A, 0x00);	// BASS_FREQUENCY_HIGH
	sta015WriteRegister(0x79, 0xC8);	// BASS_FREQUENCY_LOW
	sta015WriteRegister(0x72, 0x01);	// RUN
	sta015WriteRegister(0x13, 0x01);	// PLAY
	sta015WriteRegister(0x14, 0x00);	// MUTE
	sta015WriteRegister(0x46, 10);		// DLA
	sta015WriteRegister(0x48, 10);		// DRA
	
	puts_P(PSTR("OK!\r"));
	
	uint16_t avgBitRate, headL, headM, headH, error;
	
	// liest 65000*512 = ~32MB Daten und sendet sie an den MP3-Player
	for(uint32_t b=0; b<65000; b++) {
		
		
		///////////////////////////////////////////////////////
		//                                                   //
		//    Hier Code zum Lesen der Daten von einem        //
		//    Speichermedium einfgen                        //
		//                                                   //
		//    readData(buffer);                              //
		//                                                   //
		///////////////////////////////////////////////////////
		
		
		// alle 8kB nenn bissel Info lesen
		if((b & (0x0000000F)) == 0x0F) {
			PORTB |= (1<<PB7);
			avgBitRate = (uint16_t)sta015ReadRegister(0x6A)<<1;
			headH = (uint16_t)sta015ReadRegister(0x43);
			headM = (uint16_t)sta015ReadRegister(0x44);
			headL = (uint16_t)sta015ReadRegister(0x45);
			error = sta015ReadRegister(0x0F);
			printf_P(PSTR("%u %.2X %.2X%.2X%.2X\r"), avgBitRate, error, headH, headL, headM);
			PORTB &= ~(1<<PB7);
		}
		
		// BIT_EN auf High
		PORTB |= (1<<PB0);
		
		int16_t i=0;
		// Block an den Decoder senden
		while(i < 512) {
			// Auf Data Request warten
			while(!(PINB & (1<<PB4)));
			
			// So lange wie Data Request an ist, Daten senden
			while((PINB & (1<<PB4)) && i<512) {
				// Zeichen an den Decoder senden
				SPDR = buffer[i++];
				// auf SPI bertragung warten
				while(!(SPSR & (1<<SPIF)));
			}
		}
		
		// BIT_EN auf LOW
		PORTB &= ~(1<<PB0);
	}
	
	puts_P(PSTR("\rFertig!\r"));
	
	// HEAD lesen
	data = sta015ReadRegister(0x43);
	printf_P(PSTR("HEAD_H: 0x%.2X\r"), data);
	data = sta015ReadRegister(0x44);
	printf_P(PSTR("HEAD_M: 0x%.2X\r"), data);
	// ERROR_CODE lesen
	data = sta015ReadRegister(0x0F);
	printf_P(PSTR("ERROR_CODE: 0x%.2X\r"), data);
	
	// Endlosschleife
	while(1);
}
