#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <stdio.h>
#include <inttypes.h>

#include "uart.h"

static char uartRxBuffer[UART_RX_BUFFER_SIZE];
volatile uint8_t uartRxRdIndex = 0;
volatile uint8_t uartRxWrIndex = 0;
volatile uint8_t uartRxDataLength = 0;

static char uartTxBuffer[UART_TX_BUFFER_SIZE];
volatile uint8_t uartTxRdIndex = 0;
volatile uint8_t uartTxWrIndex = 0;
volatile uint8_t uartTxDataLength = 0;

FILE* uartIO;
volatile uint8_t uartFlags;

uint16_t uartTimeout;

/**
 * UART Empfangs-Interrupt
 */
SIGNAL(UART_RECV_IRQ) {
	char c = UART_UDR;
	
	if(uartFlags & (1 << UART_ECHO) && uartTxDataLength < UART_TX_BUFFER_SIZE) {
		uartPut(c);
	}
	
	// Empfangsbuffer voll?
	if(uartRxDataLength >= UART_RX_BUFFER_SIZE-1) {
		// Overflow-Flag setzen
		uartFlags |= 1 << UART_BUF_OV;
	
	// bertragung zu Ende?
	} else if (c != '\r') {
		// Sonst Zeichen speichern
		uartRxBuffer[uartRxWrIndex++] = c;
		// ggf. Schreib-Index umbrechen
		if(uartRxWrIndex >= UART_RX_BUFFER_SIZE) {
			uartRxWrIndex = 0;
		}
		// Daten-Counter inkrementieren
		uartRxDataLength++;
	}
	
	// Nachricht zu Ende?
	if(c == '\r') {
		// String-Terminator, ans Ende.
		uartRxBuffer[uartRxWrIndex] = 0;
		// Receive Flag setzen
		uartFlags |= 1 << UART_RECV;
	}
}

/**
 * (Ext-)UART Sende-Interrupt
 */
SIGNAL(UART_TRANS_IRQ) {
	// Weitere zu Sendende Zeichen im Buffer?
	if(uartTxDataLength) {
		// nchstes Zeichen senden
		UART_UDR = uartTxBuffer[uartTxRdIndex++];
		// ggf. Lese-Index umbrechen
		if(uartTxRdIndex >= UART_TX_BUFFER_SIZE) {
			uartTxRdIndex = 0;
		}
		// Daten-Counter dekrementieren
		uartTxDataLength--;
	} else {
		// UDR Interrupt deaktivieren
		UART_UCSRB &= ~(1 << UDRIE);
	}
}

/**
 * Gibt ein Zeichen ber das externe UART aus. Das Zeichen wird entweder
 * sofort gesendet, oder wird im Buffer gespeichert wenn sich bereits zu
 * Sendende Zeichen im Buffer befinden.
 */
int uartPut(char c) {
	// Bei vollem Buffer warten bis wieder Platz ist.
	if(uartTxDataLength >= UART_TX_BUFFER_SIZE) {
		/*date_t toTime = rtcTime;
		if(uartTimeout) {
			rtcDateAdd(&toTime, RTC_SUB_SEC, uartTimeout);
		}*/
		
		// warten, solange Buffer voll
		while(uartTxDataLength >= UART_TX_BUFFER_SIZE) {
			// Watchdog Timer zurcksetzen
			wdt_reset();
			
			// Bei Timeout Misserfolg zurckgeben
			/*if(uartTimeout && rtcCompToSystime(&toTime) < 0) {
				return -1;
			}*/
		}
	}
	
	// Interrupts deaktivieren
	cli();
	// Hardware- und Software-Buffer leer?
	if((UART_UCSRA & (1 << UDRE)) && (uartTxDataLength == 0)) {
		// Schreibvorgang initiieren
		UART_UDR = c;
	} else {
		// Sonst Zeichen im Buffer speichern
		uartTxBuffer[uartTxWrIndex++] = c;
		// ggf. Schreib-Index umbrechen
		if(uartTxWrIndex >= UART_TX_BUFFER_SIZE) {
			uartTxWrIndex = 0;
		}
		// Daten-Counter inkrementieren
		uartTxDataLength++;
	}
	// Date Register Empty Interrupt aktivieren
	UART_UCSRB |= (1 << UDRIE);
	// Interrupts wieder aktivieren
	sei();
	
	// Erfolg zurckgeben
	return 0;
}

/**
 * Die Funktion gibt ein empfangenes Zeichen zurck. Falls sich Zeichen im
 * Empfangsbuffer befinden werden zunchst diese zurckgegeben. Ist der Buffer
 * leer wird auf ein Zeichen gewartet.
 */
int uartGet(void) {
	int c;
	
	// Bei leerem Buffer auf Zeichen warten
	if(!uartRxDataLength) {
		/*date_t toTime = rtcTime;
		if(uartTimeout) {
			rtcDateAdd(&toTime, RTC_SUB_SEC, uartTimeout);
		}*/
		
		// Auf Zeichen warten
		while(!uartRxDataLength) {
			// Watchdog Timer zurcksetzen
			wdt_reset();
			
			// Bei Timeout Misserfolg zurckgeben
			/*if(uartTimeout && rtcCompToSystime(&toTime) < 0) {
				return -1;
			}*/
		}
	}
	
	// Interrupts deaktivieren
	cli();
	// Zeichen vom Buffer holen
	c = uartRxBuffer[uartRxRdIndex++];
	// ggf. Schreib-Index umbrechen
	if(uartRxRdIndex >= UART_RX_BUFFER_SIZE) {
		uartRxRdIndex = 0;
	}
	// Zhler dekrementieren
	uartRxDataLength--;
	// Interrupts wieder aktivieren
	sei();
	
	return c;
}

/**
 * Verwirdt den Inhalt des Empfangsbuffers.
 */
void uartFlushRxBuffer(void) {
	cli();
	uartRxDataLength = 0;
	uartRxRdIndex = uartRxWrIndex;
	sei();
}

/**
 * Verwirdt den Inhalt des Sendebuffers.
 */
void uartFlushTxBuffer(void) {
	cli();
	uartTxDataLength = 0;
	uartTxRdIndex = uartTxWrIndex;
	sei();
}

/**
 * Initialisiert das Externe UART (UART1).
 * Es werden die RX Complete und Data Register Empty Interrupts verwendet.
 * Baudrate, Empfangs- und Sendebuffer Gre werden im Headerfile
 * definiert.
 */
void uartInit(void) {
	// Hardware initialisieren
	UART_UCSRA = 0;
	UART_UCSRB = (1 << RXCIE) |		// Receiver IRQ Enable
				(1 << RXEN) |		// Receiver Enable
				(1 << TXEN);		// Transmitter Enable
	UART_UCSRC = 0x06;				// 8 Datenbits, 1 Stoppbit, No parity
	UART_UBRR = F_OSC/(16*UART_BAUDRATE)-1;
	
	uartFlags |= (1 << UART_ECHO);	// Echo enabled
	uartTimeout = 0;				// Timeout deaktiviert
	
	sei();							// Global Interrupt Enable
	
	// IO-Stream fr Nutzung mit stdio ffnen
	uartIO = fdevopen(uartPut, uartGet, 0);
}

/**
 * Deaktiviert das UART.
 */
void uartDisable(void) {
	// Hardware abschalten
	UART_UCSRB = 0;
}
